from abc import ABC
import logging
from ..exceptions import ModelException
from ..redis_db import POPOTO_REDIS_DB
import msgpack

import msgpack_numpy as m
m.patch()
logger = logging.getLogger(__name__)


class PublisherException(ModelException):
    pass


class PublisherModel(ABC):
    def __init__(self, *args, **kwargs):
        self.publish_data = None
        self.channel_name = kwargs.get('channel_name', self.__class__.__name__)

    def publish(self, data=None, pipeline=None, *args, **kwargs):
        # logger.debug(f"publish to {self.channel_name}: {publish_data}")
        self.publish_data = data or self.publish_data
        if not self.publish_data:
            return
        if not self.channel_name:
            raise PublisherException("missing channel to publish to")

        if pipeline:
            return pipeline.publish(self.__class__.__name__, json.dumps(self.get_z_add_data()))
        else:
            subscriber_count = POPOTO_REDIS_DB.publish(self.channel_name, msgpack.dumps(publish_data))
            logger.debug(f"published data to `{self.channel_name}`, {subscriber_count} subscribers")



from settings import LOAD_TALIB

if LOAD_TALIB:
    import talib

from apps.TA import HORIZONS
from apps.TA.storages.abstract.indicator import IndicatorStorage
from apps.TA.storages.abstract.indicator_subscriber import IndicatorSubscriber
from apps.TA.storages.data.price import PriceStorage
from settings import logger


class MidpriceStorage(IndicatorStorage):

    def produce_signal(self):
        pass


class MidpriceSubscriber(IndicatorSubscriber):
    classes_subscribing_to = [
        PriceStorage
    ]

    def handle(self, channel, data, *args, **kwargs):

        self.index = self.key_suffix

        if self.index is not 'low_price':
            logger.debug(f'index {self.index} is not `high_price` or `low_price` ...ignoring...')
            return

        new_midprice_storage = MidpriceStorage(ticker=self.ticker,
                                     exchange=self.exchange,
                                     timestamp=self.timestamp)

        for horizon in HORIZONS:
            periods = horizon * 14

            high_value_np_array = new_midprice_storage.get_denoted_price_array("high_price", periods)
            low_value_np_array = new_midprice_storage.get_denoted_price_array("low_price", periods)

            timeperiod = min([len(high_value_np_array), len(low_value_np_array), periods])
            midprice_value = talib.MIDPRICE(high_value_np_array, low_value_np_array, timeperiod=timeperiod)[-1]
            # logger.debug(f'savingMidprice value {midprice_value}for {self.ticker} on {periods} periods')

            new_midprice_storage.periods = periods
            new_midprice_storage.value = float(midprice_value)
            new_midprice_storage.save()



import logging
from abc import ABC

from ..redis_db import POPOTO_REDIS_DB
from ..exceptions import ModelException

logger = logging.getLogger(__name__)


class KeyValueException(ModelException):
    pass


class KeyValueModel(ABC):
    """
    stores things in redis database given a key and value
    by default uses the instance class name as the key
    recommend to uniquely identify the instance with a key prefix or suffix
    prefixes are for more specific categories of objects (eg. mammal:human:woman:Lisa )
    suffixes are for specific attributes (eg. Lisa:eye_color, Lisa:age, etc)
    """

    def __init__(self, *args, **kwargs):
        # key for redis storage
        self._db_key_main = kwargs.get('key', self.__class__.__name__)
        self._db_key_prefix = kwargs.get('key_prefix', "")
        self._db_key_suffix = kwargs.get('key_suffix', "")
        self._db_key = self.build_db_key()
        self.value = kwargs.get('value', "")
        self.force_save = kwargs.get('force_save', False)

    def __str__(self):
        return str(self._db_key)

    @classmethod
    def format_db_key(cls, key_main: str, key_prefix: str, key_suffix: str) -> str:
        key_main = key_main or cls.__name__
        return str(
            f'{key_prefix.strip(":")}:' +
            f'{key_main.strip(":")}' +
            f':{key_suffix.strip(":")}'
        ).replace("::", ":").strip(":")

    @classmethod
    def get(cls, db_key: str, instance_key: str = None):
        key_prefix, key_suffix = db_key.split(instance_key or cls.__name__)
        key_prefix = key_prefix.strip(":")
        key_suffix = key_suffix.strip(":")
        return cls.__new__(key=instance_key or cls.__name__, key_prefix=key_prefix, key_suffix=key_suffix)

    def build_db_key(self):
        self._db_key = str(
            self.format_db_key(
                key_main=self._db_key_main,
                key_prefix=self._db_key_prefix,
                key_suffix=self._db_key_suffix
            )
            # todo: add this line for using env in key
            # + f':{SIMULATED_ENV if SIMULATED_ENV != "PRODUCTION" else ""}'
        )
        return self._db_key

    def save(self, pipeline=None, *args, **kwargs):
        if not self.value:
            raise KeyValueException("no value set, nothing to save!")
        if not self.force_save:
            # validate some rules here?
            pass
        # logger.debug(f'savingkey, value: {self.db_key}, {self.value}')
        return POPOTO_REDIS_DB.set(self._db_key, self.value)

    def get_value(self, db_key: str = "", *args, **kwargs):
        return POPOTO_REDIS_DB.get(db_key or self._db_key)



# class ModelBase(type):
#     """Metaclass for all models."""
#     def __new__(cls, name, bases, attrs, **kwargs):
#         contributable_attrs = {}
#         for obj_name, obj in attrs.items():
#             if _has_contribute_to_class(obj):
#                 contributable_attrs[obj_name] = obj
#
#



import logging

from ..models.key_value import KeyValueModel
from ..models.publisher import PublisherModel
from ..redis_db import POPOTO_REDIS_DB, BEGINNING_OF_TIME
from ..exceptions import ModelException

logger = logging.getLogger(__name__)


class VolumeseriesException(ModelException):
    pass


class VolumeseriesModel(KeyValueModel):
    """
    stores things in a sorted set unique to each ticker and publisher
    ordered by blocks of volume instead of time
    """
    class_describer = "volumeseries"



from ..models.key_value import KeyValueModel
from ..models.graph_node import GraphNodeModel
from ..models.publisher import PublisherModel
from ..models.subscriber import SubscriberModel
from ..models.timeseries import TimeseriesModel
from ..models.volumeseries import VolumeseriesModel
